from __future__ import annotations

from typing import Optional
from typing import List
from typing import Dict
from logging import Logger
from logging import getLogger
from abc import ABC
from abc import abstractmethod
from enum import Enum
from urllib.parse import urlparse
from urllib.parse import parse_qs
from re import match

from pydantic import BaseModel
from pydantic import Field
from lxml import etree

from .session import Session

def format_size(size: int) -> str:
    units = ['KB', 'MB', 'GB', 'TB', 'PB']
    for index, unit in enumerate(units, start=1):
        if size / 1024 ** index < 1024:
            return f'{size / 1024 ** index:.2f} {unit}'
    return 'Error Size'

def get_headers(header_file_path: str) -> Dict[str, str]:
    with open(header_file_path, 'r', encoding='utf8') as file:
        headers = {}
        for line in file.readlines()[1:]:
            key, value = line.strip().split(': ', 1)
            headers[key] = value
        return headers

def get_parameters_from_href(href: str) -> Dict[str, List[str]]:
    return parse_qs(urlparse(href).query)

def get_id_from_href(href: str) -> Optional[str]:
    return get_parameters_from_href(href).get('id', [None])[0]

def convert_to_bytes(size: str) -> Optional[int]:
    result = match(r"([0-9\. ]+)([a-zA-Z]+)", size.strip())
    if not result:
        return None

    groups = result.groups()
    if not len(groups) == 2:
        return None

    number, unit = groups
    return calculate_bytes(number, unit)

def calculate_bytes(number: str, unit: str) -> int:
    units: Dict[str, int] = {
        'kb': 1024,
        'mb': 1024 ** 2,
        'gb': 1024 ** 3,
        'tb': 1024 ** 4,
        'pb': 1024 ** 5,
        'kib': 1024,
        'mib': 1024 ** 2,
        'gib': 1024 ** 3,
        'tib': 1024 ** 4,
        'pib': 1024 ** 5
    }

    return int(float(number) * units.get(unit.lower(), 1))

def find_element(html: etree._Element, xpath: str) -> Optional[etree._Element]: # pylint: disable=c-extension-no-member
    elements = html.xpath(xpath)
    return elements[0] if elements else None

class Status(str, Enum):
    LEECHING = 'leeching'
    SEEDING = 'seeding'

class Task(BaseModel):
    torrent_id: str
    torrent_name: str
    status: Status

class User(BaseModel):
    user_id: str
    user_name: str
    upload_bytes: int
    download_bytes: int
    email: Optional[str]
    bonus: float
    passkey: Optional[str] = Field(default=None)

class Promotion(BaseModel):
    upload_ratio: float
    download_ratio: float

class Torrent(BaseModel):
    torrent_id: str
    torrent_name: str
    size: int
    seeders: int
    leechers: int
    hit_and_run: int
    promotion: Promotion
    crawler: Crawler

    class Config:
        arbitrary_types_allowed = True

    def save(self, file_path: str) -> None:
        self.crawler.download_torrent(self.torrent_id, file_path)

    def __repr__(self) -> str:
        return f'<Torrent {self.crawler.__class__.__name__} {self.torrent_id} {self.torrent_name} up/down ' \
            f'{self.seeders}/{self.leechers} promotion {self.promotion.upload_ratio}/{self.promotion.download_ratio} {format_size(self.size)}>'

class Crawler(ABC):
    base_url = ''

    def __init__(
        self,
        headers: Dict[str, str],
        base_url: str = '',
        proxy: Optional[str] = None,
        logger: Optional[Logger] = None,
        qps: float = 1,
        timeout: Optional[float] = None
    ) -> None:
        self.base_url = base_url or self.base_url
        self.headers = headers
        self.proxy = proxy
        self.proxies = {'http': proxy, 'https': proxy} if proxy else {}
        self.logger = logger or getLogger('dummy')
        self.qps = qps
        self.timeout = timeout

        self.session = Session(qps=qps)
        self.session.proxies.update(self.proxies)
        self.session.headers.update(self.headers)

        self.number_pattern = r'\d+([\,]\d+)*([\.]\d+)'
        self.unit_pattern = r'KB|MB|GB|TB|PB|KiB|MiB|GiB|TiB|PiB'

    def __repr__(self) -> str:
        return f'<Crawler {self.__class__.__name__} {self.base_url} proxy: {self.proxy}, qps: {self.qps}>'

    @abstractmethod
    def get_user(self) -> User:
        return NotImplemented

    @abstractmethod
    def get_torrents(self, pages: int = 1) -> List[Torrent]:
        return NotImplemented

    @abstractmethod
    def get_torrent(self, torrent_id: str) -> Torrent:
        return NotImplemented

    @abstractmethod
    def download_torrent(self, torrent_id: str, file_path: str) -> bool:
        return NotImplemented

    @abstractmethod
    def get_tasks(self) -> List[Task]:
        return NotImplemented

Torrent.update_forward_refs()



from re import match
from typing import List
from typing import Dict
from typing import Optional
from http import HTTPStatus

from lxml import etree # pylint: disable=c-extension-no-member

from pydantic import ValidationError

from .base import Crawler
from .base import User
from .base import Torrent
from .base import Task
from .base import Promotion
from .base import get_id_from_href
from .base import find_element
from .base import Status
from .base import calculate_bytes
from .base import convert_to_bytes
from .exceptions import CannotGetUserInformationException
from .exceptions import CannotGetTorrentInformationException
from .exceptions import RequestException

def get_promotion(element: Optional[etree._Element]) -> Promotion: # pylint: disable=c-extension-no-member
    if element is None:
        return Promotion(upload_ratio=1, download_ratio=1)

    clazz = element.get('class')
    if not clazz:
        return Promotion(upload_ratio=1, download_ratio=1)

    if clazz == 'pro_free':
        return Promotion(upload_ratio=1, download_ratio=0)

    if clazz == 'pro_50pctdown':
        return Promotion(upload_ratio=1, download_ratio=0.5)

    if clazz == 'pro_30pctdown':
        return Promotion(upload_ratio=1, download_ratio=0.3)

    return Promotion(upload_ratio=1, download_ratio=1)

class CHDBits(Crawler):
    base_url: str = 'https://chdbits.xyz'
    hr_policy: Dict[str, int] = {
        'h3': 3 * 24 * 3600,
        'h5': 5 * 24 * 3600,
        '3day': 3 * 24 * 3600,
        '5day': 5 * 24 * 3600,
    }

    def get_user(self) -> User:
        pattern = r'[\s\S]*'.join(
            [
                r'欢迎回来, (?P<user_name>.+)  \[退出\]',
                fr'\[使用\]: (?P<bonus>{self.number_pattern}) 邀请',
                fr'上传量： (?P<upload>{self.number_pattern}) (?P<upload_unit>{self.unit_pattern})',
                fr'下载量： (?P<download>{self.number_pattern}) (?P<download_unit>{self.unit_pattern})'
            ]
        )

        response = self.session.get(url=self.base_url + '/usercp.php')
        if not response.status_code == HTTPStatus.OK:
            raise RequestException(response)

        html = etree.HTML(response.text) # pylint: disable=c-extension-no-member

        email_element = find_element(html, '//*[@id="outer"]/table[2]/tr[2]/td[2]')
        user_id_element = find_element(html, '//*[@id="info_block"]/tr/td/table/tr/td[1]/span/span/a')
        title_element = find_element(html, '//*[@id="info_block"]/tr/td/table/tr/td[1]/span')
        passkey_element = find_element(html, '//*[@id="outer"]/table[2]/tr[4]/td[2]')

        if email_element is None or user_id_element is None or title_element is None or passkey_element is None:
            raise CannotGetUserInformationException()

        result = match(pattern, ''.join(title_element.itertext()))
        if not result:
            raise CannotGetUserInformationException()

        user_id = get_id_from_href(user_id_element.get('href'))
        if not user_id:
            raise CannotGetUserInformationException()

        return User(
            user_id=user_id,
            user_name=result.group('user_name'),
            upload_bytes=calculate_bytes(result.group('upload'), result.group('upload_unit')),
            download_bytes=calculate_bytes(result.group('download'), result.group('download_unit')),
            email=email_element.text,
            bonus=float(result.group('bonus').replace(',', '')),
            passkey=passkey_element.text
        )

    def get_torrents(self, pages: int = 1) -> List[Torrent]:
        torrents = []
        for page in range(pages):
            response = self.session.get(
                url=self.base_url + '/torrents.php',
                params={'page': str(page), 'incldead': '0', 'spstate': '0'}
            )

            if not response.status_code == HTTPStatus.OK:
                raise RequestException(response)

            html = etree.HTML(response.text) # pylint: disable=c-extension-no-member
            _, *rows = html.xpath('/html/body/table[2]/tr[2]/td/table/tr/td/table/tr')

            for row in rows:
                title_element = find_element(row, 'td[2]/table/tr/td[1]/a')
                size_element = find_element(row, 'td[5]')
                seeders_element = find_element(row, 'td[6]')
                leechers_element = find_element(row, 'td[7]')
                hit_and_run_element = find_element(row, './/div[@class="circle-text"]')
                promotion_element = find_element(row, './/img[starts-with(@class, "pro_")]')

                if title_element is None or size_element is None or seeders_element is None or leechers_element is None:
                    self.logger.warning(CannotGetTorrentInformationException())
                    continue

                torrent_id = get_id_from_href(title_element.get('href'))
                size = convert_to_bytes(' '.join(size_element.itertext()))

                if not torrent_id or not size:
                    self.logger.warning(CannotGetTorrentInformationException())
                    continue

                try:
                    torrent = Torrent(
                        torrent_id=torrent_id,
                        torrent_name=''.join(title_element.itertext()),
                        size=size,
                        seeders=int(''.join(seeders_element.itertext()).replace(',', '')),
                        leechers=int(''.join(leechers_element.itertext()).replace(',', '')),
                        hit_and_run=self.hr_policy.get(hit_and_run_element.text, 0) if hit_and_run_element is not None else 0,
                        promotion=get_promotion(promotion_element),
                        crawler=self,
                    )
                except ValidationError as exception:
                    self.logger.warning(exception)
                else:
                    torrents.append(torrent)

        return torrents

    def get_torrent(self, torrent_id: str) -> Torrent:
        response = self.session.get(
            url=self.base_url + '/details.php',
            params={'hit': '1', 'id': torrent_id}
        )

        if not response.status_code == HTTPStatus.OK:
            raise RequestException(response)

        html = etree.HTML(response.text) # pylint: disable=c-extension-no-member

        title_element = find_element(html, '//*[@id="top"]')
        base_information_element = find_element(html, '//*[@id="outer"]/table[1]/tr[3]/td[2]')
        seeder_and_leecher_element = find_element(html, '//*[@id="peercount"]')
        promotion_element = find_element(title_element, './/img[starts-with(@class, "pro_")]')

        if title_element is None or base_information_element is None or seeder_and_leecher_element is None:
            raise CannotGetTorrentInformationException()

        result = match(r'大小：(?P<size_number>.+)类型', ''.join(base_information_element.itertext()))
        if not result:
            raise CannotGetTorrentInformationException()
        size = convert_to_bytes(result.group('size_number').strip())

        result = match(r'[\s\S]*H&R:(?P<hit_and_run_string>.+)', ''.join(base_information_element.itertext()))
        if not result:
            hit_and_run = 0
        else:
            hit_and_run = self.hr_policy.get(result.group('hit_and_run_string').strip(), 0)

        torrent_name = title_element.text.strip()

        result = match(r'(?P<seeder_number>.+)个做种者 \| (?P<leecher_number>.+)个下载者', ''.join(seeder_and_leecher_element.itertext()))

        if not result or not size:
            raise CannotGetTorrentInformationException()

        return Torrent(
            torrent_id=torrent_id,
            torrent_name=torrent_name,
            size=size,
            seeders=int(result.group('seeder_number')),
            leechers=int(result.group('leecher_number')),
            promotion=get_promotion(promotion_element),
            crawler=self,
            hit_and_run=hit_and_run
        )

    def download_torrent(self, torrent_id: str, file_path: str) -> bool:
        response = self.session.get(
            url=self.base_url + '/download.php',
            params={'id': torrent_id}
        )

        if not response.status_code == HTTPStatus.OK:
            self.logger.warning(RequestException(response))
            return False

        with open(file_path, 'wb') as file:
            file.write(response.content)
            return True

    def get_tasks(self) -> List[Task]:
        user = self.get_user()
        tasks = []

        for status in [Status.LEECHING, Status.SEEDING]:
            response = self.session.get(
                url=self.base_url + '/getusertorrentlistajax.php',
                params={'userid': user.user_id, 'type': status}
            )

            if not response.status_code == HTTPStatus.OK:
                self.logger.warning(RequestException(response))
                continue

            html = etree.HTML(response.text) # pylint: disable = c-extension-no-member

            _, *rows = html.xpath('//tr')
            for row in rows:
                title_element = find_element(row, 'td[2]/a')
                href = title_element.get('href') or '' if title_element is not None else ''
                torrent_id = get_id_from_href(href)
                torrent_name = title_element.get('title') if title_element is not None else ''

                if not torrent_name or not torrent_id:
                    self.logger.warning('2333')
                    continue

                tasks.append(Task(torrent_id=torrent_id, torrent_name=torrent_name, status=status))

        return tasks



from math import inf
from time import time
from time import sleep

from requests import Session as BaseSession
from requests.adapters import HTTPAdapter as BaseHTTPAdapter


class HTTPAdapter(BaseHTTPAdapter):
    def __init__(self, timeout: int, *args, **kwargs): # type: ignore
        self.timeout = timeout
        super().__init__(*args, **kwargs)

    def send(self, *args, **kwargs): # type: ignore
        kwargs['timeout'] = kwargs.get('timeout', self.timeout)
        return super().send(*args, **kwargs)

class Session(BaseSession):
    def __init__(self, *args, qps: float = inf, timeout: int = 1, **kwargs): # type: ignore
        super().__init__(*args, **kwargs)
        self.mount('http://', HTTPAdapter(timeout=timeout))
        self.mount('https://', HTTPAdapter(timeout=timeout))
        self.interval = 1 / qps
        self.last_request_time = time() - self.interval

    def request(self, *args,**kwargs): # type: ignore
        sleep_time = max(self.last_request_time + self.interval - time(), 0)
        sleep(sleep_time)

        response = super().request(*args, **kwargs)
        self.last_request_time = time()
        return response



from re import match
from typing import List
from typing import Optional
from http import HTTPStatus

from lxml import etree # pylint: disable=c-extension-no-member

from pydantic import ValidationError

from .base import Crawler
from .base import User
from .base import Torrent
from .base import Task
from .base import Promotion
from .base import get_id_from_href
from .base import find_element
from .base import Status
from .base import calculate_bytes
from .base import convert_to_bytes
from .exceptions import CannotGetUserInformationException
from .exceptions import CannotGetTorrentInformationException
from .exceptions import RequestException

def get_promotion(element: Optional[etree._Element]) -> Promotion: # pylint: disable=c-extension-no-member
    src = element.get('src') if element is not None else None

    if src == '/pic/ico_free.gif':
        return Promotion(upload_ratio=1, download_ratio=0)

    if src == '/pic/ico_30.gif':
        return Promotion(upload_ratio=1, download_ratio=0.3)

    if src == '/pic/ico_half.gif':
        return Promotion(upload_ratio=1, download_ratio=0.5)

    return Promotion(upload_ratio=1, download_ratio=1)
class TTG(Crawler):
    base_url: str = 'https://totheglory.im/'

    def get_user(self) -> User:
        pattern = r'[\s\S]*'.join(
            [
                '',
                rf'上传量 :  (?P<upload_number>{self.number_pattern}) (?P<upload_unit>{self.unit_pattern})\xa0\xa0',
                rf'下载量 : (?P<download_number>{self.number_pattern}) (?P<download_unit>{self.unit_pattern})\xa0\xa0',
                rf'积分 : (?P<bonus>{self.number_pattern})\xa0\xa0'
            ]
        )

        response = self.session.get(
            url=self.base_url + '/my.php',
        )

        html = etree.HTML(response.content.decode('utf8')) # pylint: disable=c-extension-no-member

        email_element = find_element(html, '//*[@id="main_table"]/tr[1]/td/table/tr[2]/td/form/table/tr[5]/td[2]')
        passkey_element = find_element(html, '//*[@id="main_table"]/tr[1]/td/table/tr[2]/td/form/table/tr[6]/td[2]')
        title_element = find_element(html, '/html/body/table[2]/tr/td/table/tr/td[1]')
        user_id_element = find_element(html, '//*[@id="main_table"]/tr[1]/td/h1/a')

        if email_element is None or user_id_element is None or title_element is None or passkey_element is None:
            raise CannotGetUserInformationException()

        result = match(pattern, ''.join(title_element.itertext()))
        if not result:
            raise CannotGetUserInformationException()

        user_id = get_id_from_href(user_id_element.get('href'))
        if not user_id:
            raise CannotGetUserInformationException()

        return User(
            user_id=user_id,
            user_name=user_id_element.text,
            upload_bytes=calculate_bytes(result.group('upload_number'), result.group('upload_unit')),
            download_bytes=calculate_bytes(result.group('download_number'), result.group('download_unit')),
            email=email_element.text,
            bonus=float(result.group('bonus')),
            passkey=passkey_element.text
        )

    def get_torrents(self, pages: int = 1) -> List[Torrent]:
        torrents = []

        for page in range(pages):
            response = self.session.get(url=self.base_url + '/browse.php', params={'c': 'M', 'page': str(page)})
            if not response.status_code == HTTPStatus.OK:
                raise RequestException(response)

            html = etree.HTML(response.content.decode('utf8')) # pylint: disable=c-extension-no-member
            _, *rows = html.xpath('/html/body/table[3]/tr[1]/td/form/table/tr')
            for row in rows:
                torrent_id_element = find_element(row, './td[2]/div[1]/a')
                title_element = find_element(row, './td[2]/div[1]/a/b')
                size_element = find_element(row, './td[7]')
                seeders_and_leechers_element = find_element(row, './td[9]')

                if torrent_id_element is None or title_element is None or size_element is None or seeders_and_leechers_element is None:
                    self.logger.warning(CannotGetTorrentInformationException())
                    continue

                torrent_id_result = match(r'/t/(?P<torrent_id>.+)/', torrent_id_element.get('href') or '')
                size = convert_to_bytes(' '.join(size_element.itertext()))

                seeders_and_leechers_result = match(r'(?P<seeders>\d+)/\n(?P<leechers>\d+)', ''.join(seeders_and_leechers_element.itertext()))
                if not size or not torrent_id_result or not seeders_and_leechers_result or torrent_id_element is None or seeders_and_leechers_element is None:
                    self.logger.warning(CannotGetTorrentInformationException())
                    continue

                promotion_element = find_element(row, './/img[starts-with(@src, "/pic/ico_")]')
                hit_and_run_element = find_element(row, './/img[@title="Hit and Run"]')

                try:
                    torrent = Torrent(
                        torrent_id=torrent_id_result.group('torrent_id'),
                        torrent_name=title_element.text,
                        size=size,
                        seeders=seeders_and_leechers_result.group('seeders'),
                        leechers=seeders_and_leechers_result.group('leechers'),
                        hit_and_run=60 * 3600 if hit_and_run_element is not None else 0,
                        promotion=get_promotion(promotion_element),
                        crawler=self,
                    )
                except ValidationError as exception:
                    self.logger.warning(exception)
                    continue
                else:
                    torrents.append(torrent)

        return torrents

    def get_torrent(self, torrent_id: str) -> Torrent:
        response = self.session.get(self.base_url + f'/t/{torrent_id}/')
        if not response.status_code == HTTPStatus.OK:
            raise RequestException(response)

        html = etree.HTML(response.content.decode('utf8')) # pylint: disable=c-extension-no-member

        title_element = find_element(html, '//*[@id="main_table"]/tr[1]/td/h1')
        size_element = find_element(html, "//td[preceding-sibling::td[1][text()='尺寸']]")
        seeders_and_leechers_element = find_element(html, "//td[preceding-sibling::td[1][text()='活跃用户']]")

        if title_element is None or size_element is None or seeders_and_leechers_element is None:
            raise CannotGetTorrentInformationException()

        size_result = match(rf'(?P<size_number>{self.number_pattern}) (?P<size_unit>{self.unit_pattern})', size_element.text)
        seeders_and_leechers_result = match(r'(?P<seeders>.+) 做种者，(?P<leechers>.+) 下载者', ''.join(seeders_and_leechers_element.itertext()))
        if not size_result or not seeders_and_leechers_result:
            raise CannotGetTorrentInformationException()

        hit_and_run_element = find_element(html, './/img[@alt="Hit & Run"]')
        promotion_element = find_element(html, './/img[starts-with(@src, "/pic/ico_")]')

        return Torrent(
            torrent_id=torrent_id,
            torrent_name=title_element.text,
            size=calculate_bytes(size_result.group('size_number'), size_result.group('size_unit')),
            seeders=seeders_and_leechers_result.group('seeders'),
            leechers=seeders_and_leechers_result.group('leechers'),
            hit_and_run=60 * 3600 if hit_and_run_element is not None else 0,
            promotion=get_promotion(promotion_element),
            crawler=self,
        )

    def download_torrent(self, torrent_id: str, file_path: str) -> bool:
        response = self.session.get(self.base_url + f'/t/{torrent_id}/')
        if not response.status_code == HTTPStatus.OK:
            self.logger.warning(RequestException(response))
            return False

        html = etree.HTML(response.content.decode('utf8')) # pylint: disable=c-extension-no-member
        download_url_element = find_element(html, '//*[@id="main_table"]/tr[1]/td/table[1]/tr[1]/td[2]/a[1]')
        if download_url_element is None:
            self.logger.warning(CannotGetTorrentInformationException())
            return False

        href = download_url_element.get('href')
        if download_url_element is None:
            self.logger.warning(CannotGetTorrentInformationException())
            return False

        response = self.session.get(self.base_url + href)
        if not response.status_code == HTTPStatus.OK:
            self.logger.warning(RequestException(response))
            return False

        with open(file_path, 'wb') as file:
            file.write(response.content)
            return True

    def get_tasks(self) -> List[Task]:
        user = self.get_user()

        response = self.session.get(self.base_url + '/userdetails.php', params={'id': user.user_id})
        if not response.status_code == HTTPStatus.OK:
            raise RequestException(response)

        html = etree.HTML(response.content.decode('utf8')) # pylint: disable=c-extension-no-member

        tasks = []
        for keyword, status in [('当前上传', Status.SEEDING), ('当前下载', Status.LEECHING)]:
            try:
                _, *rows = html.xpath(f'''//td[preceding-sibling::td[1][text()={repr(keyword)}]]//tr''')
            except ValueError:
                continue

            for row in rows:
                title_element = find_element(row, './/td[2]/a/b')
                link_element = find_element(row, './/td[2]/a')

                if title_element is None or link_element is None:
                    self.logger.warning(CannotGetTorrentInformationException())
                    continue

                torrent_id = get_id_from_href(self.base_url + link_element.get('href', '/'))
                tasks.append(Task(torrent_id=torrent_id, torrent_name=title_element.text, status=status))
        return tasks



from typing import List
from http import HTTPStatus

from pydantic import BaseModel
from pydantic import Field
from pydantic import validator
from pydantic import ValidationError

from .base import Crawler
from .base import Torrent
from .base import Status
from .base import User
from .base import Task
from .base import convert_to_bytes
from .base import Promotion
from .exceptions import RequestException

class Peers(BaseModel):
    upload: int
    download: int

class TorrentStatus(BaseModel):
    download_ratio: float = Field(alias='downCoefficient')
    upload_ratio: float = Field(alias='upCoefficient')

class TaskItem(BaseModel):
    torrent_id: str = Field(alias='tid')
    torrent_name: str = Field(alias='title')

class TaskData(BaseModel):
    items: List[TaskItem] = Field(alias='list')
    total_pages: int = Field(alias='maxPage')

class ListTaskResponse(BaseModel):
    data: TaskData

class TorrentItem(BaseModel):
    torrent_id: str = Field(alias='tid')
    torrent_name: str = Field(alias='title')
    size: str = Field(alias='fileSize')
    peers: Peers
    promotion: TorrentStatus = Field(alias='status')

    @validator('size')
    def size_validator(cls, value: str) -> int:
        result = convert_to_bytes(value)
        if not result:
            raise ValueError(f'Invalid size {value}')
        return result

    @validator('promotion')
    def promotion_validator(cls, value: TorrentStatus) -> Promotion:
        return Promotion(upload_ratio=value.upload_ratio, download_ratio=value.download_ratio)

class TorrentList(BaseModel):
    items: List[TorrentItem] = Field(alias='list')

class ListTorrentsResponse(BaseModel):
    data: TorrentList

class UserData(BaseModel):
    upload_bytes: int = Field(alias='upload')
    download_bytes: int = Field(alias='download')
    user_id: str = Field(alias='uid', coerce_numbers_to_str=True)
    user_name: str = Field(alias='username')
    bonus: float = Field(alias='point')
    passkey: str

class GetUserResponse(BaseModel):
    data: UserData

class GetTorrentData(BaseModel):
    torrent: TorrentItem

class GetTorrentResponse(BaseModel):
    data: GetTorrentData

class FSM(Crawler):
    base_url: str = 'https://api.fsm.name'

    def get_torrents(self, pages: int = 1) -> List[Torrent]:
        torrents = []

        for page in range(pages):
            response = self.session.get(
                url=self.base_url + '/Torrents/listTorrents',
                params={
                    'type': '0',
                    'systematics': '0',
                    'tags': '[]',
                    'keyword': '',
                    'page': str(page + 1)
                }
            )

            if not response.status_code == HTTPStatus.OK:
                raise RequestException(response)

            list_torrents_response = ListTorrentsResponse.parse_obj(response.json())
            for item in list_torrents_response.data.items:
                try:
                    torrent = Torrent(
                        torrent_id=item.torrent_id,
                        torrent_name=item.torrent_name,
                        size=item.size,
                        hit_and_run=False,
                        promotion=item.promotion,
                        seeders=item.peers.upload,
                        leechers=item.peers.download,
                        crawler=self
                    )
                except ValidationError as exception:
                    self.logger.warning(exception)
                    continue
                else:
                    torrents.append(torrent)

        return torrents

    def get_user(self) -> User:
        response = self.session.get(self.base_url + '/Users/infos')
        if not response.status_code == HTTPStatus.OK:
            raise RequestException(response)

        get_user_response: GetUserResponse = GetUserResponse.parse_obj(response.json())

        return User(
            user_id=get_user_response.data.user_id,
            user_name=get_user_response.data.user_name,
            email=None,
            upload_bytes=get_user_response.data.upload_bytes,
            download_bytes=get_user_response.data.download_bytes,
            bonus=get_user_response.data.bonus,
            passkey=get_user_response.data.passkey
        )

    def get_torrent(self, torrent_id: str) -> Torrent:
        response = self.session.get(
            url=self.base_url + '/Torrents/details',
            params={'tid': torrent_id}
        )

        if not response.status_code == HTTPStatus.OK:
            raise RequestException(response)

        get_torrent_response: GetTorrentResponse = GetTorrentResponse.parse_obj(response.json())
        item = get_torrent_response.data.torrent

        return Torrent(
            torrent_id=item.torrent_id,
            torrent_name=item.torrent_name,
            size=item.size,
            hit_and_run=False,
            promotion=item.promotion,
            seeders=item.peers.upload,
            leechers=item.peers.download,
            crawler=self
        )

    def download_torrent(self, torrent_id: str, file_path: str) -> bool:
        user = self.get_user()

        response = self.session.get(
            url=self.base_url + '/Torrents/download',
            params={'passkey': user.passkey, 'tid': torrent_id, 'source': 'direct'}
        )

        if not response.status_code == HTTPStatus.OK:
            self.logger.warning(RequestException(response))
            return False

        with open(file_path, 'wb') as file:
            file.write(response.content)
            return True

    def get_tasks(self) -> List[Task]:
        page = 1
        tasks = []
        for api_path, status in [('/Torrents/listMyDownload', Status.LEECHING), ('/Torrents/listMySeed', Status.SEEDING)]:
            while True:
                response = self.session.get(self.base_url + api_path, params={'page': str(page)})

                if not response.status_code == HTTPStatus.OK:
                    self.logger.warning(RequestException(response))

                list_tasks_response: ListTaskResponse = ListTaskResponse.parse_obj(response.json())

                for item in list_tasks_response.data.items:
                    tasks.append(Task(torrent_id=item.torrent_id, torrent_name=item.torrent_name, status=status))

                if page >= list_tasks_response.data.total_pages:
                    break

                page += 1
        return tasks



from re import match
from typing import List
from typing import Optional
from http import HTTPStatus

from lxml import etree # pylint: disable=c-extension-no-member

from pydantic import ValidationError

from .base import Crawler
from .base import User
from .base import Torrent
from .base import Task
from .base import Promotion
from .base import get_id_from_href
from .base import find_element
from .base import Status
from .base import calculate_bytes
from .base import convert_to_bytes
from .exceptions import CannotGetUserInformationException
from .exceptions import CannotGetTorrentInformationException
from .exceptions import RequestException

promotion_map = {
    'free': Promotion(upload_ratio=1, download_ratio=0),
    'twoup': Promotion(upload_ratio=2, download_ratio=1),
    'twoupfree': Promotion(upload_ratio=2, download_ratio=0),
    'halfdown': Promotion(upload_ratio=1, download_ratio=0.5),
    'twouphalfdown': Promotion(upload_ratio=2, download_ratio=0.5),
    'thirtypercent': Promotion(upload_ratio=1, download_ratio=0.3),
}

def get_promotion_from_list(element: Optional[etree._Element]) -> Promotion: # pylint: disable=c-extension-no-member
    clazz = element.get('class') if element is not None else 'normal'
    if clazz == 'pro_free':
        return Promotion(upload_ratio=1, download_ratio=0)
    if clazz == 'pro_50pctdown':
        return Promotion(upload_ratio=1, download_ratio=0.5)
    if clazz == 'pro_2up':
        return Promotion(upload_ratio=2, download_ratio=1)
    if clazz == 'pro_free2up':
        return Promotion(upload_ratio=2, download_ratio=0)
    if clazz == 'pro_50pctdown2up':
        return Promotion(upload_ratio=2, download_ratio=0.5)
    if clazz == 'pro_30pctdown':
        return Promotion(upload_ratio=1, download_ratio=0.3)
    return Promotion(upload_ratio=1, download_ratio=1)

def get_promotion_from_detail(element: Optional[etree._Element]) -> Promotion: # pylint: disable=c-extension-no-member
    clazz = element.get('class') if element is not None else 'normal'
    return promotion_map.get(clazz, Promotion(upload_ratio=1, download_ratio=1))

class OurBits(Crawler):
    base_url: str = 'https://ourbits.club'

    def get_user(self) -> User:
        pattern = r'.*'.join(
            [
                r'欢迎回来, (?P<user_name>.+) \[退出\]',
                rf'魔力值 \[使用\]: (?P<bonus>{self.number_pattern}) ',
                rf'上传量： (?P<upload_number>{self.number_pattern}) (?P<upload_unit>{self.unit_pattern}) ',
                rf'下载量： (?P<download_number>{self.number_pattern}) (?P<download_unit>{self.unit_pattern}) '
                ''
            ]
        )

        response = self.session.get(url=self.base_url + '/usercp.php', timeout=self.timeout)
        if not response.status_code == HTTPStatus.OK:
            raise RequestException(response)

        html = etree.HTML(response.text) # pylint: disable=c-extension-no-member

        email_element = find_element(html, "//td[preceding-sibling::td[1][text()='邮箱地址']]")
        passkey_element = find_element(html, "//td[preceding-sibling::td[1][text()='密钥']]")
        user_id_element = find_element(html, '//*[@id="info_block"]/tr/td/table/tr/td[1]/span/span[1]/a')
        title_element = find_element(html, '//*[@id="info_block"]/tr/td/table/tr/td[1]/span')

        if email_element is None or user_id_element is None or title_element is None or passkey_element is None:
            raise CannotGetUserInformationException()

        result = match(pattern, ' '.join(''.join(title_element.itertext()).split()))
        if not result:
            raise CannotGetUserInformationException()

        user_id = get_id_from_href(user_id_element.get('href'))
        if not user_id:
            raise CannotGetUserInformationException()

        return User(
            user_id=user_id,
            user_name=result.group('user_name'),
            upload_bytes=calculate_bytes(result.group('upload_number'), result.group('upload_unit')),
            download_bytes=calculate_bytes(result.group('download_number'), result.group('download_unit')),
            email=email_element.text,
            bonus=float(result.group('bonus').replace(',', '')),
            passkey=''.join(passkey_element.itertext())
        )

    def get_torrents(self, pages: int = 1) -> List[Torrent]:
        torrents = []
        for page in range(pages):
            response = self.session.get(
                url=self.base_url + '/torrents.php',
                params={'page': str(page), 'incldead': '0', 'spstate': '0'},
                timeout=self.timeout
            )

            if not response.status_code == HTTPStatus.OK:
                raise RequestException(response)

            html = etree.HTML(response.text) # pylint: disable=c-extension-no-member
            _, *rows = html.xpath('//*[@id="torrenttable"]/tr')

            for row in rows:
                title_element = find_element(row, 'td[2]/table/tr/td[1]/a')
                size_element = find_element(row, 'td[5]')
                seeders_element = find_element(row, 'td[6]')
                leechers_element = find_element(row, 'td[7]')
                promotion_element = find_element(row, './/img[starts-with(@class, "pro_")]')

                if title_element is None or size_element is None or seeders_element is None or leechers_element is None:
                    self.logger.warning(CannotGetTorrentInformationException())
                    continue

                torrent_id = get_id_from_href(title_element.get('href'))
                size = convert_to_bytes(' '.join(size_element.itertext()))

                if not torrent_id or not size:
                    self.logger.warning(CannotGetTorrentInformationException())
                    continue

                try:
                    torrent = Torrent(
                        torrent_id=torrent_id,
                        torrent_name=''.join(title_element.itertext()).strip(),
                        size=size,
                        seeders=int(''.join(seeders_element.itertext()).replace(',', '')),
                        leechers=int(''.join(leechers_element.itertext()).replace(',', '')),
                        hit_and_run=False,
                        promotion=get_promotion_from_list(promotion_element),
                        crawler=self,
                    )
                except ValidationError as exception:
                    self.logger.warning(exception)
                else:
                    torrents.append(torrent)

        return torrents

    def get_torrent(self, torrent_id: str) -> Torrent:
        response = self.session.get(
            url=self.base_url + '/details.php',
            params={'hit': '1', 'id': torrent_id}
        )

        if not response.status_code == HTTPStatus.OK:
            raise RequestException(response)

        html = etree.HTML(response.text) # pylint: disable=c-extension-no-member

        title_element = find_element(html, '//*[@id="top"]')
        size_element = find_element(html, "//td[preceding-sibling::td[1][text()='基本信息']]")
        seeder_and_leecher_element = find_element(html, "//td[preceding-sibling::td[1][text()='同伴']]")
        promotion_element = find_element(title_element, './/font[@class="free" or @class="twoup" or @class="twoupfree" or @class="halfdown" or @class="twouphalfdown" or @class="thirtypercent"]')

        if title_element is None or size_element is None or seeder_and_leecher_element is None:
            raise CannotGetTorrentInformationException()

        result = match('大小：(?P<size>.+) 类型', ' '.join(''.join(size_element.itertext()).split()))
        if not result:
            raise CannotGetTorrentInformationException()

        size = convert_to_bytes(result.group('size'))
        torrent_name = title_element.text.strip()

        result = match(r'(?P<seeder_number>.+)个做种者\|(?P<leecher_number>.+)个下载者', ''.join(item.strip() for item in list(seeder_and_leecher_element.itertext())))
        if not result or not size:
            raise CannotGetTorrentInformationException()

        return Torrent(
            torrent_id=torrent_id,
            torrent_name=torrent_name,
            size=size,
            seeders=int(result.group('seeder_number')),
            leechers=int(result.group('leecher_number')),
            promotion=get_promotion_from_detail(promotion_element),
            crawler=self,
            hit_and_run=False
        )

    def download_torrent(self, torrent_id: str, file_path: str) -> bool:
        response = self.session.get(
            url=self.base_url + '/download.php',
            params={'id': torrent_id}
        )

        if not response.status_code == HTTPStatus.OK:
            self.logger.warning(RequestException(response))
            return False

        with open(file_path, 'wb') as file:
            file.write(response.content)
            return True

    def get_tasks(self) -> List[Task]:
        user = self.get_user()
        tasks = []

        for status in [Status.LEECHING, Status.SEEDING]:
            response = self.session.get( # pylint: disable = missing-timeout
                url=self.base_url + '/getusertorrentlistajax.php',
                params={'userid': user.user_id, 'type': status},
                timeout=self.timeout
            )

            if not response.status_code == HTTPStatus.OK:
                self.logger.warning(RequestException(response))
                continue

            html = etree.HTML(response.text) # pylint: disable = c-extension-no-member

            try:
                _, *rows = html.xpath('//tr')
            except ValueError:
                continue

            for row in rows:
                title_element = find_element(row, 'td[2]/a')
                href = title_element.get('href') or '' if title_element is not None else ''
                torrent_id = get_id_from_href(href)
                torrent_name = title_element.get('title') if title_element is not None else ''

                if not torrent_name or not torrent_id:
                    self.logger.warning('2333')
                    continue

                tasks.append(Task(torrent_id=torrent_id, torrent_name=torrent_name, status=status))

        return tasks



#!/usr/bin/env python3
# coding: utf-8

from setuptools import setup
from setuptools import find_packages

from crawlers import VERSION

with open('crawlers/requirements.txt', 'r', encoding='utf8') as file:
    install_requires = list(map(lambda x: x.strip(), file.readlines()))

with open('readme.md', 'r', encoding='utf8') as file:
    long_description = file.read()

setup(
    name='pt-crawler',
    version=VERSION,
    author='kinopico',
    author_email='zqmillet@qq.com',
    url='https://github.com/zqmillet/pt-crawler',
    description='a group of crawlers for private tracker website',
    packages=find_packages(),
    install_requires=install_requires,
    long_description=long_description,
    long_description_content_type='text/markdown',
    package_data={"crawlers": ["py.typed"]}
)
